@@ -1,7 +1,4 @@
-from pathlib import Path
-
 import click
-from appdirs import user_data_dir
 from dunamai import Version, Style
 
 from pynfogen import __version__
@@ -15,20 +12,17 @@ from pynfogen.cli.template import template
     help_option_names=["-?", "-h", "--help"],
     max_content_width=116  # max PEP8 line-width, -4 to adjust for initial indent
 ))
-@click.pass_context
-def cli(ctx):
-    user_dir = Path(user_data_dir("pynfogen", "PHOENiX"))
-    config_path = user_dir / "config.yml"
-    ctx.obj = {
-        "user_dir": user_dir,
-        "config_path": config_path,
-        "templates": user_dir / "templates",
-        "artwork": user_dir / "artwork"
-    }
+def cli() -> None:
+    """
+    \b
+    pynfogen
+    Scriptable MediaInfo-fed NFO Generator for Movies and TV.
+    https://github.com/rlaphoenix/pynfogen
+    """
 
 
 @cli.command()
-def about():
+def about() -> None:
     """Shows information about pynfogen."""
     print(
         "pynfogen - Python NFO Generator.\n"
@@ -39,7 +33,7 @@ def about():
 
 
 @cli.command()
-def version():
+def version() -> None:
     """Shows the version of the project."""
     try:
         v = Version.from_git().serialize(style=Style.SemVer)
