@@ -1,7 +1,10 @@
 import logging
+from typing import Optional
 
 import click
-import yaml
+import pytomlpp
+
+from pynfogen.config import config as data, Files
 
 
 @click.command()
@@ -10,24 +13,15 @@ import yaml
 @click.option("--unset", is_flag=True, default=False, help="Unset/remove the configuration value.")
 @click.option("--list", "list_", is_flag=True, default=False, help="List all set configuration values.")
 @click.pass_context
-def config(ctx, key: str, value: str, unset: bool, list_: bool):
+def config(ctx: click.Context, key: Optional[str], value: Optional[str], unset: bool, list_: bool) -> None:
     """Manage configuration."""
     if not key and not value and not list_:
         return click.echo(config.get_help(ctx))
 
     log = logging.getLogger("config")
 
-    config_path = ctx.obj["config_path"]
-    if config_path.exists():
-        with config_path.open() as f:
-            data = yaml.safe_load(f)
-    else:
-        data = None
-    if not data:
-        data = {}
-
     if list_:
-        print(yaml.dump(data).rstrip())
+        print(pytomlpp.dumps(data).rstrip())
         return
 
     tree = key.split(".")
@@ -49,6 +43,5 @@ def config(ctx, key: str, value: str, unset: bool, list_: bool):
         else:
             temp[tree[-1]] = value
             log.info(f"Set {key} to {repr(value)}")
-            config_path.parent.mkdir(parents=True, exist_ok=True)
-            with config_path.open("wt") as f:
-                yaml.dump(data, f)
+            Files.config.parent.mkdir(parents=True, exist_ok=True)
+            pytomlpp.dump(data, Files.config)
