@@ -30,20 +30,48 @@ class CustomFormats(Formatter):
 
     @staticmethod
     def bbimg(value: Union[List[Union[dict, str]], Union[dict, str]]) -> Union[List[str], str]:
-        """Convert a list of values into a list of BBCode [LIST][IMG] strings."""
+        """
+        Convert a list of values into a list of BBCode [LIST][IMG] strings.
+        If only one item is provided, then a single BBCode string will be provided, not a list.
+
+        Example:
+            >>> f = CustomFormats()
+            >>> f.bbimg("https://source.unsplash.com/random")
+            '[URL=https://source.unsplash.com/random][IMG]https://source.unsplash.com/random[/IMG][/URL]'
+            >>> f.bbimg({'url': 'https://picsum.photos/id/237/info', 'src': 'https://picsum.photos/id/237/200/300'})
+            '[URL=https://picsum.photos/id/237/info][IMG]https://picsum.photos/id/237/200/300[/IMG][/URL]'
+            >>> f.bbimg([{'url': 'https://foo...', 'src': 'https://bar...'}, 'https://bizz...', ...])
+            ['[URL=https://foo...][IMG]https://bar...[/IMG][/URL]',
+            '[URL=https://bizz...][IMG]https://bizz...[/IMG][/URL]', ...]
+        """
         if not value:
             return ""
         if not isinstance(value, list):
             value = [value]
-        value = [({"url": x, "src": x} if not isinstance(x, dict) else x) for x in value]
-        value = [f"[URL={x['url']}][IMG]{x['src']}[/IMG][/URL]" for x in value]
-        if len(value) == 1:
-            return value[0]
-        return value
+        images = [
+            ({"url": x, "src": x} if not isinstance(x, dict) else x)
+            for x in value
+        ]
+        bb = [f"[URL={x['url']}][IMG]{x['src']}[/IMG][/URL]" for x in images]
+        if len(bb) == 1:
+            return bb[0]
+        return bb
 
     @staticmethod
     def layout(value: Union[List[str], str], width: int, height: int, spacing: int) -> str:
-        """Lay out data in a grid with specific lengths, heights, and spacing."""
+        """
+        Lay out data in a grid with specific lengths, heights, and spacing.
+
+        Example:
+            >>> f = CustomFormats()
+            >>> f.layout(['1', '2', '3', '4'], width=2, height=2, spacing=0)
+            12
+            34
+            >>> f.layout(['1', '2', '3', '4'], width=2, height=2, spacing=1)
+            1 2
+
+            3 4
+        """
         if not value:
             return ""
         if not isinstance(value, list):
@@ -51,10 +79,13 @@ class CustomFormats(Formatter):
         if len(value) != width * height:
             # TODO: How about just ignore and try fill as much as it can?
             raise ValueError("Layout invalid, not enough images...")
-        value = [(value[i:i + width]) for i in range(0, len(value), width)]
-        value = [(" " * spacing).join(x) for x in value]
-        value = ("\n" * (spacing + 1)).join(value)
-        return value
+        grid = [
+            (value[i:i + width])
+            for i in range(0, len(value), width)
+        ]
+        grid_indented = [(" " * spacing).join(x) for x in grid]
+        grid_str = ("\n" * (spacing + 1)).join(grid_indented)
+        return grid_str
 
     def wrap(self, value: Union[List[str], str], indent: int, width: int) -> str:
         """Text-wrap data at a specific width and indent amount."""
@@ -85,7 +116,7 @@ class CustomFormats(Formatter):
             return self.center(value, *map(int, format_spec[2:].split("x")))
         return super().format_field(value, format_spec)
 
-    def list_to_indented_strings(self, value: list, indent: int = 0):
+    def list_to_indented_strings(self, value: list, indent: int = 0) -> str:
         """Recursively convert a list to an indented \n separated string."""
         if isinstance(value[0], list):
             return self.list_to_indented_strings(value[0], indent)
